﻿Module BSHorizon

    '//////////////////////////////////////////////////////////////////////
    '   Appendix
    '   Nonogaki S, Masumoto S and Shiono K (2008) Optimal Determination
    '   of Geologic Boundary Surface Using Cubic B-Spline. Geoinformatics,
    '   vol.19, no.2, pp.61-77.
    '
    '   Fortran77 program for determination of an optimal geologic surface
    '   using both elevation data and strike-dip data.
    '
    '   Note: Run a command "ulimit -s unlimited" if you get stack size
    '         error in running this program.
    '
    '                                        Producted by Susumu Nonogaki
    '                                        Last update February 17, 2009
    '   1st update(version2): revise the method of knot calculation
    '   2nd update(version3): change output format of optimal surface
    '//////////////////////////////////////////////////////////////////////
    '   Basic concepts
    '   1) A surface f(x,y) is expressed by a tensor product form of Cubic
    '      B-spline functions:
    '        f(x,y) = double sigma (c(ij)Ni(x)Nj(y))
    '                              (i = 1 to mx+3; j = 1 to my+3)
    '      where Ni(x) and Nj(y) are cubic B-splines, and c(ij) are
    '      coefficients multiplied by each tensor product which may be
    '      determined in calculation.
    '   2) Location of outcrop (xh(k), yh(k), zh(k)) provides equality and
    '      inequality constraints for the elevation of surface;
    '        f(xh(k),yh(k)) = zh(k)  if lm(k) = 0
    '        f(xh(k),yh(k)) < zh(k)  if lm(k) < 0
    '        f(xh(k),yh(k)) > zh(k)  if lm(k) > 0
    '      where lm(k) is a indicator of data types.
    '   3) Strike-dip information at location (xd(k), yd(k), zd(k)) also
    '      provides constraints for the partial derivative of surface;
    '        fx(xd(k),yd(k)) + sin(tr(k)) * tan(dp(k)) = 0
    '        fy(xd(k),yd(k)) + cos(tr(k)) * tan(dp(k)) = 0
    '      where tr(k) is trend of the maximum slope measured clockwise
    '      from north and dp(k) is dip angle from horizontal surface.
    '   4) The surface determination problem is considered as a following
    '      constrained optimization problem:
    '        Find an optimal solution f(x,y) that minimizes objective
    '        function J(f) which evaluates the smoothness of the surface;
    '          J(f) = m1 * J1(f) + m2 * J2(f)
    '               = m1 * [double integral {(fx)**2 + (fy)**2} dxdy]
    '               + m2 * [double integral {(fxx)**2+2*(fxy)**2+(fyy)**2}dxdy]
    '        subject to
    '          f(xh(k),yh(k)) = zh(k) ; if lm(k) = 0
    '          f(xh(k),yh(k)) < zh(k) ; if lm(k) < 0
    '          f(xh(k),yh(k)) > zh(k) ; if lm(k) > 0
    '        and
    '          f(xd(k),yd(k)) + sin(tr(k))*tan(dp(k)) = 0
    '          f(xd(k),yd(k)) + cos(tr(k))*tan(dp(k)) = 0.
    '   5) To solve the problem, an augmented objective function:
    '        Q(f;alpha) = J(f) + alpha * R(f)
    '      is introduced based on the exterior penalty function method. Where
    '      R(f), defined in a form of residual mean of squares, evaluates
    '      the goodness of fit and alpha is a constant called a penalty.
    '      The optimal surface is a solution of linear simultaneous equations:
    '        A c = b
    '      derived from dQ/dc(ij) = 0. The equations are solved by Choleski's method.
    '   6) The optimal surface can be saved as the form of both function and
    '      grid data. In the case of function output, the coefficients c(ij),
    '      which give the optimal surface, are saved. In grid data output,
    '      a set of values f(ig,jg) is estimated by assigning the normalized
    '      coordinates of grid points (xg(ig),yg(jg)) (ig = 1 to nxg; jg = 1 to nyg)
    '      to the equation of surface. 


    Dim A(,) As Double, B() As Double
    Dim c() As Double

    Dim neq, nbs As Long
    Dim hx, hy As Double


    Sub solve(ByRef f(,) As Double, ByVal mx As Integer, ByVal my As Integer, ByVal Nx As Long, ByVal Ny As Long, _
    ByVal Xmin As Double, ByVal Ymin As Double, ByVal Xmax As Double, ByVal Ymax As Double, _
    ByVal Dx As Double, ByVal Dy As Double, _
    ByVal Xh() As Double, ByVal Yh() As Double, ByVal Zh() As Double, _
    ByVal Lm() As Integer, ByVal Nh As Long, _
    ByVal Xd() As Double, ByVal Yd() As Double, ByVal Dfdx() As Double, ByVal Dfdy() As Double, _
    ByVal Nd As Long, _
    ByVal Pm1 As Double, ByVal Pm2 As Double, ByVal Gam As Double, _
    ByVal AlpMin As Double, ByVal AlpMax As Double, _
    ByVal Nitr As Long)

        '/////////////////subroutine solve/////////////////////////////////////
        '   Solve the simultaneous equations based on the exterior penalty method.
        '      Q(f;alpha) = J(smoothness) + alpha * R(residual) ---> min
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      hx,hy      : knot interval
        '      c(neq)     : solution (coefficients of approximation)
        '   Variables
        '      alpmin,alpmax : initial and final penalty in iterative calculation
        '      nitr          : number of iteration
        '      gam           : weight of trend & dip data to elevation data
        '      pm1,pm2       : weight of 1st & 2nd order partial derivative term in J
        '      neq           : number of rows    of matrix A (number of equations)
        '      nbs           : number of columns of matrix A (for choleski's method)
        '      A(neq,nbs)    : coefficient matrix of simultaneous equations
        '      b(neq)        : constant matrix of simultaneous equations
        '//////////////////////////////////////////////////////////////////////


        Dim Rn As Double
        Dim Alp As Double
        Dim Ratio As Double
        Dim Nhdt As Long, Nddt As Long

        Dim QRhav, QRdav As Double
        Dim QJx, QJy, QJxx, QJxy, QJyy, QJ1, QJ2 As Double

        Dim QR, QRh, QRd As Double
        Dim aR As Double
        Dim Q, Qj As Double

        Dim bx0(4), by0(4), bx1(4), by1(4) As Double
        Dim Xg(Nx), Yg(Ny) As Double
        Dim ig, jg As Long
        Dim ipx, ipy As Long
        Dim rx, ry As Double
        Dim ij As Long

        Static i, j As Long, Itr As Long

        Dim itry As Integer

        '---------Penalty : alpha for elevation data---------------------------
        itry = 0
        For i = 1 To Nh
            itry = itry + Lm(i) ^ 2
        Next
        '
        If itry = 0 Then
            Ratio = 1
            Nitr = 1
        Else
            Rn = 1 / (Nitr - 1)
            Ratio = (AlpMax / AlpMin) ^ Rn
        End If

        '---------Penalty : gamma for trend & dip data-------------------------
        If Nd = 0 Then
            Gam = 0
        End If

        '---------Set knot interval--------------------------------------------
        hx = (Xmax - Xmin) / mx
        hy = (Ymax - Ymin) / my

        '---------Calculate the number of rows & columns of matrix A-----------
        neq = (mx + 3) * (my + 3)
        nbs = (mx + 3) * 3 + 4

        ReDim A(neq, nbs), B(neq), c(neq)

        '---------Initialize matrix c------------------------------------------
        For i = 1 To neq
            c(i) = 0
        Next

        '---------Iterative calculation----------------------------------------
        For Itr = 1 To Nitr

            For i = 1 To neq
                B(i) = 0
                For j = 1 To nbs
                    A(i, j) = 0
                Next
            Next
            Alp = AlpMin * (Ratio ^ (Itr - 1))

            '---------Set matrix b and Ah------------------------------------------
            setbAh(mx, my, Xh, Yh, Zh, Lm, Nh, Xmin, Xmax, Ymin, Ymax, Alp, Nhdt)


            '---------Set matrix b and Ad------------------------------------------
            setbAd(mx, my, Xd, Yd, Dfdx, Dfdy, Nd, Xmin, Xmax, Ymin, Ymax, Alp, Gam, Nddt)

            '---------Set matrix Ax,Ay,Axx,Axy and Ayy-----------------------------
            setJ(mx, my, Pm1, Pm2)

            '---------Solve simultaneous equations---------------------------------
            choles()

            '---------Evaluate Rh--------------------------------------------------
            Rhvalue(mx, my, Xh, Yh, Zh, Lm, Nh, Xmin, Xmax, Ymin, Ymax, Nhdt, QRh, QRhav)

            '---------Evaluate Rd--------------------------------------------------
            Rdvalue(mx, my, Xd, Yd, Dfdx, Dfdy, Nd, Xmin, Xmax, Ymin, Ymax, Nddt, QRd, QRdav)

            '---------Evaluate J---------------------------------------------------
            Jvalue(mx, my, Pm1, Pm2, QJx, QJy, QJxx, QJxy, QJyy, QJ1, QJ2, Qj)

            '---------Evaluate Q---------------------------------------------------
            QR = QRh + Gam * QRd
            aR = Alp * QR
            Q = Qj + aR

        Next

        '  999 write(*,99)
        '   99 format("Error: invalid parameter input")
        '      stop
        'エラー表示なし
        ''


        '---------Calculate the height value f(ig,jg)--------------------------
        For i = 1 To Nx
            Xg(i) = Xmin + (i - 1) * Dx
        Next
        For j = 1 To Ny
            Yg(j) = Ymin + (j - 1) * Dy
        Next

        For ig = 1 To Nx
            For jg = 1 To Ny
                ipx = Int((Xg(ig) - Xmin) / hx) + 1
                ipy = Int((Yg(jg) - Ymin) / hy) + 1
                If ipx = mx + 1 Then ipx = mx
                If ipy = my + 1 Then ipy = my
                rx = (Xg(ig) - Xmin) / hx - (ipx - 1)
                ry = (Yg(jg) - Ymin) / hy - (ipy - 1)
                bspl(rx, hx, 0, bx0)
                bspl(ry, hy, 0, by0)
                f(ig, jg) = 0.0

                For j = 0 To 3
                    For i = 0 To 3

                        ij = ipx + i + (ipy - 1 + j) * (mx + 3)
                        f(ig, jg) = f(ig, jg) + c(ij) * bx0(4 - i) * by0(4 - j)

                    Next
                Next

            Next
        Next

    End Sub
    '
    Private Sub bspl(ByVal r As Double, ByVal h As Double, ByVal ider As Long, ByVal bn() As Double)
        '/////////////////subroutine bspl//////////////////////////////////////
        '   Calculate 4 non-zero B-spline at data point.
        '      4 basic functions:
        '        bn1(r) =       r**3                         / 6
        '        bn2(r) = (-3 * r**3 + 3 * r**2 + 3 * r + 1) / 6
        '        bn3(r) = ( 3 * r**3 - 6 * r**2         + 4) / 6
        '        bn4(r) = (-    r**3 + 3 * r**2 - 3 * r + 1) / 6
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      r             :  normalized coordinate (0.0 to 1.0)
        '      h             :  knot interval
        '      ider          :  order of derivative
        '      bn(4)         :  4 non-zero B-spline
        '//////////////////////////////////////////////////////////////////////

        Dim At As Double

        If ider = 0 Then
            At = 1.0 / 6.0
            bn(1) = r ^ 3 * At
            bn(2) = (3.0 * (-r ^ 3 + r ^ 2 + r) + 1.0) * At
            bn(3) = (3.0 * r ^ 3 - 6.0 * r ^ 2 + 4.0) * At
            bn(4) = (-r ^ 3 + 3.0 * (r ^ 2 - r) + 1.0) * At
        ElseIf ider = 1 Then
            At = 0.5 / h
            bn(1) = r ^ 2 * At
            bn(2) = (-3.0 * r ^ 2 + 2.0 * r + 1.0) * At
            bn(3) = (3.0 * r ^ 2 - 4.0 * r) * At
            bn(4) = (-r ^ 2 + 2.0 * r - 1.0) * At
        ElseIf ider = 2 Then
            At = 1.0 / (h ^ 2)
            bn(1) = r * At
            bn(2) = (-3.0 * r + 1.0) * At
            bn(3) = (3.0 * r - 2.0) * At
            bn(4) = (-r + 1.0) * At
        End If

    End Sub
    '

    Private Sub setbAh(ByVal mx As Integer, ByVal my As Integer, ByVal xh() As Double, ByVal yh() As Double, ByVal zh() As Double, ByVal lm() As Integer, ByVal nh As Long, _
                       ByVal xmin As Double, ByVal xmax As Double, ByVal ymin As Double, ByVal ymax As Double, _
                       ByVal alp As Double, ByVal nhdt As Long)

        '/////////////////subroutine setbAh////////////////////////////////////
        '   Set the elements of matrix b & A (Ah) derived from elevation data.
        '      Ah(ij,i'j') = alpha / nh * sigma (Nx(i)Nx(i')Ny(j)Ny(j'))
        '      b(ij)       = alpha / nh * sigma (Nx(i)Ny(j)zh)
        '      (i,i' = 1 to mx + 3; j,j' = 1 to my + 3)
        '   When knot number on right-hand neighbor is 'i', only 4 B-spline
        '   have non-zero values, that is, N(i), N(i+1), N(i+2) and N(i+3).
        '   Using 4 non-zero B-spline, N(i+j) can be expressed as follows.
        '      N(ip+j) = b(4-j)    (j = 0 to 3)
        '----------------------------------------------------------------------
        '   Variables
        '      ip* : knot number on right-hand neighbor
        '      r*  : normalized coordinate from knot on left-hand neigbor
        '//////////////////////////////////////////////////////////////////////

        Dim bx0(4), by0(4), bx1(4), by1(4) As Double

        Dim fh As Double
        Dim wt As Double
        Dim rx, ry As Double
        Dim hres As Double

        Dim id As Long
        Dim ipx, ipy As Long

        Dim i, j, ij, i1, j1, i2, j2, ia, ja As Long
        '
        nhdt = 0

        For id = 1 To nh
            If xh(id) >= xmin And xh(id) <= xmax And yh(id) >= ymin And yh(id) <= ymax Then nhdt = nhdt + 1
        Next

        wt = alp / nhdt

        For id = 1 To nh
            If xh(id) >= xmin And xh(id) <= xmax And yh(id) >= ymin And yh(id) <= ymax Then

                '--Determine knot number on right-hand neighbor------------------------
                ipx = Int((xh(id) - xmin) / hx) + 1
                ipy = Int((yh(id) - ymin) / hy) + 1

                '--Consider the data on final knot as the one in the final section-----
                If ipx = mx + 1 Then ipx = mx
                If ipy = my + 1 Then ipy = my

                '--Calcurate normalized coordinate from knot on left-hand neigbor------
                rx = (xh(id) - xmin) / hx - (ipx - 1)
                ry = (yh(id) - ymin) / hy - (ipy - 1)

                '--Calculate 4 non-zero B-spline with respect to x & y-----------------
                bspl(rx, hx, 0, bx0)
                bspl(ry, hy, 0, by0)

                '--Examine whether data is satisfied with constraints or not-----------
                fh = 0.0

                For j = 0 To 3
                    For i = 0 To 3
                        ij = ipx + i + (ipy - 1 + j) * (mx + 3)
                        fh = fh + c(ij) * bx0(4 - i) * by0(4 - j)
                    Next
                Next

                hres = fh - zh(id)

                If lm(id) = 0 Or hres * lm(id) <= 0 Then

                    '--Set b & Ah using 4 non-zero B-spline--------------------------------
                    For j1 = 0 To 3
                        For i1 = 0 To 3
                            ia = ipx + i1 + (ipy - 1 + j1) * (mx + 3)
                            B(ia) = B(ia) + wt * bx0(4 - i1) * by0(4 - j1) * zh(id)

                            For j2 = 0 To 3
                                For i2 = 0 To 3
                                    ja = i2 - i1 + 1 + (j2 - j1) * (mx + 3)
                                    If ja > 0 Then A(ia, ja) = A(ia, ja) + wt * bx0(4 - i1) * by0(4 - j1) * bx0(4 - i2) * by0(4 - j2)
                                Next
                            Next
                        Next
                    Next

                End If
            End If

        Next

    End Sub

    Private Sub setbAd(ByVal mx As Integer, ByVal my As Integer, ByVal Xd() As Double, ByVal Yd() As Double, _
                       ByVal Dfdx() As Double, ByVal Dfdy() As Double, ByVal Nd As Long, _
                       ByVal Xmin As Double, ByVal Xmax As Double, ByVal Ymin As Double, ByVal Ymax As Double, _
                       ByVal Alp As Double, ByVal Gam As Double, ByVal Nddt As Long)

        '/////////////////subroutine setbAd////////////////////////////////////
        '    Set the elements of matrix b and A (Ad) derived from trend & dip data.
        '       Ad(ij,i'j') =   alpha * gamma / nd 
        '         * sigma (Nx'(i)Nx'(i')Ny(j)Ny(j')+Nx(i)Nx(i')Ny'(j)Ny'(j'))
        '       b(ij)       = - alpha * gamma / nd
        '         * sigma (Nx'(i)Ny(j)sin(tr)tan(dp)+Nx(i)Ny'(j)cos(tr)tan(dp))
        '       (i,i' = 1 to mx + 3; j,j' = 1 to my + 3)
        '//////////////////////////////////////////////////////////////////////


        Dim bx0(4), by0(4), bx1(4), by1(4) As Double
        Dim beta, wt As Double
        Dim rx, ry As Double

        Static id As Long
        Dim ipx, ipy As Long

        Dim i1, j1, i2, j2, ia, ja As Long

        Nddt = 0

        If Nd <> 0 Then

            For id = 1 To Nd
                If Xd(id) >= Xmin And Xd(id) <= Xmax And Yd(id) >= Ymin And Yd(id) <= Ymax Then Nddt = Nddt + 1
            Next

            beta = Alp * Gam
            wt = beta / Nddt
            For id = 1 To Nd
                If Xd(id) >= Xmin And Xd(id) <= Xmax And Yd(id) >= Ymin And Yd(id) <= Ymax Then

                    '--Determine knot number on right-hand neighbor------------------------
                    ipx = Int((Xd(id) - Xmin) / hx) + 1
                    ipy = Int((Yd(id) - Ymin) / hy) + 1

                    '--Consider the data on final knot as the one in the final section-----
                    If ipx = mx + 1 Then ipx = mx
                    If ipy = my + 1 Then ipy = my

                    '--Calcurate normalized coordinate from knot on left-hand neigbor------
                    rx = (Xd(id) - Xmin) / hx - (ipx - 1)
                    ry = (Yd(id) - Ymin) / hy - (ipy - 1)

                    '--Calculate 4 non-zero B-spline with respect to x & y-----------------
                    bspl(rx, hx, 0, bx0)
                    bspl(rx, hx, 1, bx1)
                    bspl(ry, hy, 0, by0)
                    bspl(ry, hy, 1, by1)

                    '--Set b & Ad using 4 non-zero B-spline--------------------------------
                    For j1 = 0 To 3
                        For i1 = 0 To 3
                            ia = ipx + i1 + (ipy - 1 + j1) * (mx + 3)
                            B(ia) = B(ia) + wt * bx1(4 - i1) * by0(4 - j1) * Dfdx(id) + wt * bx0(4 - i1) * by1(4 - j1) * Dfdy(id)
                            For j2 = 0 To 3
                                For i2 = 0 To 3
                                    ja = i2 - i1 + 1 + (j2 - j1) * (mx + 3)
                                    If ja > 0 Then A(ia, ja) = A(ia, ja) + wt * bx1(4 - i1) * by0(4 - j1) * bx1(4 - i2) * by0(4 - j2) + wt * bx0(4 - i1) * by1(4 - j1) * bx0(4 - i2) * by1(4 - j2)
                                Next
                            Next
                        Next
                    Next

                End If
            Next
        End If

    End Sub
    '

    Private Sub setJ(ByVal mx As Integer, ByVal my As Integer, ByVal pm1 As Double, ByVal pm2 As Double)

        '/////////////////subroutine setJ//////////////////////////////////////
        '    Set the elements of matrix A (Ax, Ay, Axx, Ayy and Axy) derived
        '    from the integration of 1st & 2nd order partial derivative squared
        '    with respect to x & y.
        '       Ax(ij,i'j')  =     m1 * [integral Nx'(i)Nx'(i') dx]
        '                                       * [integral Ny(j)Ny(j') dy]
        '       Ay(ij,i'j')  =     m1 * [integral Nx(i)Nx(i') dx]
        '                                       * [integral Ny'(j)Ny'(j') dy]
        '       Axx(ij,i'j') =     m2 * [integral Nx''(i)Nx''(i') dx]
        '                                       * [integral Ny(j)Ny(j') dy]
        '       Axy(ij,i'j') = 2 * m2 * [integral Nx'(i)Nx'(i') dx]
        '                                       * [integral Ny'(j)Ny'(j') dy]
        '       Ayy(ij,i'j') =     m2 * [integral Nx(i)Nx(i') dx]
        '                                       * [integral Ny''(j)Ny''(j') dy]
        '       (i,i' = 1 to mx + 3; j,j' = 1 to my + 3)
        '//////////////////////////////////////////////////////////////////////

        Dim ffix0(4), ffix1(4), ffix2(4) As Double
        Dim ffiy0(4), ffiy1(4), ffiy2(4) As Double

        Dim wx, wy, wxx, wxy, wyy As Double

        Dim ia, i22, j22 As Long
        Static i1, j1 As Long

        wx = pm1 * (hy / hx) / (mx * My)
        wy = pm1 * (hx / hy) / (mx * My)
        wxx = pm2 * hy / (hx ^ 3)
        wxy = 2.0 * pm2 / (hx * hy)
        wyy = pm2 * hx / (hy ^ 3)

        For j1 = 1 To my + 3

            '--Calculate the integration with respect to y-------------------------
            integral(my, 0, j1, ffiy0)
            integral(my, 1, j1, ffiy1)
            integral(my, 2, j1, ffiy2)
            j22 = 4

            '        if(j1.ge.my+1) j22=my+4-j1
            If j1 >= my + 1 Then j22 = my + 4 - j1

            For i1 = 1 To mx + 3

                '--Calculate the integration with respect to x-------------------------
                integral(mx, 0, i1, ffix0)
                integral(mx, 1, i1, ffix1)
                integral(mx, 2, i1, ffix2)
                ia = i1 + (j1 - 1) * (mx + 3)
                i22 = 4
                If i1 >= mx + 1 Then i22 = mx + 4 - i1

                '--Set Ax, Ay, Axx, Ayy and Axy----------------------------------------
                setJtmp(mx, wx, ia, i22, j22, ffix1, ffiy0)
                setJtmp(mx, wy, ia, i22, j22, ffix0, ffiy1)
                setJtmp(mx, wxx, ia, i22, j22, ffix2, ffiy0)
                setJtmp(mx, wxy, ia, i22, j22, ffix1, ffiy1)
                setJtmp(mx, wyy, ia, i22, j22, ffix0, ffiy2)
            Next
        Next

    End Sub

    Private Sub setJtmp(ByVal mx As Integer, ByVal w As Double, ByVal ia As Long, ByVal i22 As Long, ByVal j22 As Long, ByRef ffix() As Double, ByRef ffiy() As Double)

        '/////////////////subroutine setJtmp///////////////////////////////////
        '    Calculate the individual element values of A by multiply weight
        '    with two kinds of integration.
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      w         : weight of element 
        '      ia        : row number of objective element
        '      i22,j22   : indicator wheather element may be copyed or not
        '      ffix,ffiy : integration value
        '//////////////////////////////////////////////////////////////////////

        Dim tmp As Double
        Static i2, j2 As Long
        Static ja, iia, jja As Long

        For j2 = 1 To j22
            For i2 = 1 To i22
                ja = i2 + (j2 - 1) * (mx + 3)
                tmp = w * ffix(i2) * ffiy(j2)
                A(ia, ja) = A(ia, ja) + tmp
                If j2 >= 2 And i2 >= 2 Then
                    iia = ia + i2 - 1
                    jja = ja - 2 * (i2 - 1)
                    A(iia, jja) = A(iia, jja) + tmp
                End If
            Next
        Next

    End Sub

    Private Sub integral(ByVal m As Long, ByVal ider As Long, ByVal ia As Long, ByRef ffi() As Double)

        '/////////////////subroutine integral//////////////////////////////////
        '   Calculate integration of 4 basic non-zero B-spline functions
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      m      : number of division of the region
        '      ider   : order of derivative
        '      ia     : knot number
        '      ffi(4) : integration value of 4 basic functions
        '//////////////////////////////////////////////////////////////////////

        '--Calculate the integration of basic functions------------------------
        If ider = 0 Then
            If ia = 1 Then
                ffi(1) = 1.0 / 252
                ffi(2) = 43 / 1680
                ffi(3) = 1.0 / 84
                ffi(4) = 1.0 / 5040
            ElseIf ia = 2 Then
                ffi(1) = 151 / 630
                ffi(2) = 59 / 280
                ffi(3) = 1.0 / 42
                ffi(4) = 1.0 / 5040
            ElseIf ia = 3 Then
                ffi(1) = 599 / 1260
                ffi(2) = 397 / 1680
                ffi(3) = 1.0 / 42
                ffi(4) = 1.0 / 5040
            ElseIf ia = m + 1 Then
                ffi(1) = 599 / 1260
                ffi(2) = 59 / 280
                ffi(3) = 1.0 / 84
                ffi(4) = 0.0
            ElseIf ia = m + 2 Then
                ffi(1) = 151 / 630
                ffi(2) = 43 / 1680
                ffi(3) = 0.0
                ffi(4) = 0.0
            ElseIf ia = m + 3 Then
                ffi(1) = 1.0 / 252
                ffi(2) = 0.0
                ffi(3) = 0.0
                ffi(4) = 0.0
            Else
                ffi(1) = 151 / 315
                ffi(2) = 397 / 1680
                ffi(3) = 1.0 / 42
                ffi(4) = 1.0 / 5040
            End If

            '--Calculate the integration of 1st order
            '         partial derivative of basic functions------------------------
        ElseIf ider = 1 Then
            If ia = 1 Then
                ffi(1) = 1.0 / 20
                ffi(2) = 7.0 / 120
                ffi(3) = -1.0 / 10
                ffi(4) = -1.0 / 120
            ElseIf ia = 2 Then
                ffi(1) = 1.0 / 3.0
                ffi(2) = -11 / 60
                ffi(3) = -1.0 / 5.0
                ffi(4) = -1.0 / 120
            ElseIf ia = 3 Then
                ffi(1) = 37 / 60
                ffi(2) = -1.0 / 8.0
                ffi(3) = -1.0 / 5.0
                ffi(4) = -1.0 / 120
            ElseIf ia = m + 1 Then
                ffi(1) = 37 / 60
                ffi(2) = -11 / 60
                ffi(3) = -1.0 / 10
                ffi(4) = 0.0
            ElseIf ia = m + 2 Then
                ffi(1) = 1.0 / 3.0
                ffi(2) = 7.0 / 120
                ffi(3) = 0.0
                ffi(4) = 0.0
            ElseIf ia = m + 3 Then
                ffi(1) = 1.0 / 20
                ffi(2) = 0.0
                ffi(3) = 0.0
                ffi(4) = 0.0
            Else
                ffi(1) = 2.0 / 3.0
                ffi(2) = -1.0 / 8.0
                ffi(3) = -1.0 / 5.0
                ffi(4) = -1.0 / 120
            End If

            '--Calculate the integration of 2nd order
            '         partial derivative of basic functions------------------------
        ElseIf ider = 2 Then
            If ia = 1 Then
                ffi(1) = 1.0 / 3.0
                ffi(2) = -1.0 / 2.0
                ffi(3) = 0.0
                ffi(4) = 1.0 / 6.0
            ElseIf ia = 2 Then
                ffi(1) = 4.0 / 3.0
                ffi(2) = -1.0
                ffi(3) = 0.0
                ffi(4) = 1.0 / 6.0
            ElseIf ia = 3 Then
                ffi(1) = 7.0 / 3.0
                ffi(2) = -3.0 / 2.0
                ffi(3) = 0.0
                ffi(4) = 1.0 / 6.0
            ElseIf ia = m + 1 Then
                ffi(1) = 7.0 / 3.0
                ffi(2) = -1.0
                ffi(3) = 0.0
                ffi(4) = 0.0
            ElseIf ia = m + 2 Then
                ffi(1) = 4.0 / 3.0
                ffi(2) = -1.0 / 2.0
                ffi(3) = 0.0
                ffi(4) = 0.0
            ElseIf ia = m + 3 Then
                ffi(1) = 1.0 / 3.0
                ffi(2) = 0.0
                ffi(3) = 0.0
                ffi(4) = 0.0
            Else
                ffi(1) = 8.0 / 3.0
                ffi(2) = -3.0 / 2.0
                ffi(3) = 0.0
                ffi(4) = 1.0 / 6.0
            End If
        End If

    End Sub

    Private Sub choles()

        '/////////////////subroutine choles////////////////////////////////////
        '    Solving a normal equation with choleski's method
        '    See Ichida & Yoshimoto (1979) for original program.
        '//////////////////////////////////////////////////////////////////////
        '

        Dim ia, ias, iat As Long
        Dim tx, tki, bki As Double
        Dim iasj, iam1, ka, kat As Long
        Dim jt, jtl, jtt As Long
        Dim neqm1, net As Long


        A(1, 1) = Math.Sqrt(A(1, 1))
        B(1) = B(1) / A(1, 1)

        For jt = 2 To nbs
            A(1, jt) = A(1, jt) / A(1, 1)
        Next

        For ia = 2 To neq

            iam1 = ia - 1
            If ia >= nbs Then
                ias = ia - nbs + 1
            Else
                ias = 1
            End If
            tki = 0.0

            For ka = ias To iam1
                iat = ia - ka + 1
                tki = tki + A(ka, iat) ^ 2
            Next

            A(ia, 1) = Math.Sqrt(A(ia, 1) - tki)

            If neq - ia >= nbs Then
                jtl = nbs
            Else
                jtl = neq - ia + 1
                If jtl = 0 Then GoTo 38
            End If

            For jt = 2 To jtl

                If ia >= nbs Then
                    iasj = ias + jt - 1
                Else
                    If jt > nbs - ia Then
                        iasj = jt - (nbs - ia)
                    Else
                        iasj = 1
                    End If
                End If
                tki = 0.0

                If iasj <= iam1 Then
                    For ka = iasj To iam1
                        iat = ia - ka + 1
                        jtt = iat + jt - 1
                        tki = tki + A(ka, iat) * A(ka, jtt)
                    Next
                End If
                A(ia, jt) = (A(ia, jt) - tki) / A(ia, 1)


            Next jt

38:         bki = 0.0
            For ka = ias To iam1
                iat = ia - ka + 1
                bki = bki + A(ka, iat) * B(ka)
            Next

            B(ia) = (B(ia) - bki) / A(ia, 1)
        Next

        c(neq) = B(neq) / A(neq, 1)
        neqm1 = neq - 1
        For iat = 1 To neqm1
            ia = neq - iat
            If iat >= nbs Then
                net = nbs
            Else
                net = iat + 1
            End If
            tx = 0

            For ka = 2 To net
                kat = ia + ka - 1
                tx = tx + A(ia, ka) * c(kat)
            Next

            c(ia) = (B(ia) - tx) / A(ia, 1)
        Next

    End Sub
    '

    Private Sub Rhvalue(ByVal mx As Integer, ByVal my As Integer, ByVal Xh() As Double, ByVal Yh() As Double, ByVal Zh() As Double, _
                        ByVal Lm() As Integer, ByVal Nh As Long, ByVal Xmin As Double, ByVal Xmax As Double, _
                        ByVal Ymin As Double, ByVal Ymax As Double, ByVal Nhdt As Long, _
                        ByVal QRh As Double, ByVal QRhav As Double)

        '/////////////////subroutine Rhvalue///////////////////////////////////
        '   Evaluate the residual mean of squares between elevation data and
        '   final solution. 
        '       Rh(f) = sigma [{double sigma (c(ij)Nx(i)Ny(j)) - zh}**2] / nh
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      QRh   : residual mean of squares
        '      QRhav : square root of QRh
        '//////////////////////////////////////////////////////////////////////

        Dim bx0(4), by0(4) As Double
        Dim hres As Double
        Dim rx, ry, rh As Double
        Dim fh As Double

        Static i, j, ij, id As Long
        Dim ipx, ipy As Long
        '
        Nhdt = 0
        rh = 0.0

        For id = 1 To Nh
            If Xh(id) >= Xmin And Xh(id) <= Xmax And Yh(id) >= Ymin And Yh(id) <= Ymax Then
                ipx = Int((Xh(id) - Xmin) / hx) + 1
                ipy = Int((Yh(id) - Ymin) / hy) + 1
                If ipx = mx + 1 Then ipx = mx
                If ipy = my + 1 Then ipy = my
                rx = (Xh(id) - Xmin) / hx - (ipx - 1)
                ry = (Yh(id) - Ymin) / hy - (ipy - 1)
                bspl(rx, hx, 0, bx0)
                bspl(ry, hy, 0, by0)
                fh = 0.0

                For j = 0 To 3
                    For i = 0 To 3
                        ij = ipx + i + (ipy - 1 + j) * (mx + 3)
                        fh = fh + c(ij) * bx0(4 - i) * by0(4 - j)
                    Next
                Next
                hres = fh - Zh(id)

                If Lm(id) = 0 Or hres * Lm(id) < 0 Then
                    rh = rh + hres ^ 2
                    Nhdt = Nhdt + 1
                End If
            End If
        Next
        QRh = rh / (Nhdt)
        QRhav = Math.Sqrt(QRh)

    End Sub

    Private Sub Rdvalue(ByVal mx As Integer, ByVal my As Integer, ByVal xd() As Double, ByVal yd() As Double, ByVal dfdx() As Double, ByVal dfdy() As Double, ByVal nd As Long, ByVal xmin As Double, ByVal xmax As Double, ByVal ymin As Double, ByVal ymax As Double, ByVal nddt As Long, ByVal QRd As Double, ByVal QRdav As Double)

        '/////////////////subroutine Rdvalue///////////////////////////////////
        '   Evaluate the residual mean of squares between trend & dip data and
        '   final solution.
        '       Rd(f) = sigma [
        '               {double sigma (c(ij)Nx'(i)Ny(j))+sin(tr)tan(dp)}**2
        '               {double sigma (c(ij)Nx(i)Ny'(j))+cos(tr)tan(dp)}**2
        '               ] / nd
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      QRd   : residual mean of squares
        '      QRdav : square root of QRd
        '//////////////////////////////////////////////////////////////////////

        Dim bx0(4), by0(4), bx1(4), by1(4) As Double

        Dim Rdx, Rdy As Double
        Dim dxres, dyres As Double
        Dim rx, ry As Double
        Dim fdx, fdy As Double

        Dim id, ipx, ipy, i, j, ij As Long

        If nd = 0 Then
            QRd = 0.0
            QRdav = 0.0
            Exit Sub
        End If

        Rdx = 0.0
        Rdy = 0.0

        For id = 1 To nd
            If xd(id) >= xmin And xd(id) <= xmax And yd(id) >= ymin And yd(id) <= ymax Then
                ipx = Int((xd(id) - xmin) / hx) + 1
                ipy = Int((yd(id) - ymin) / hy) + 1
                If ipx = mx + 1 Then ipx = mx
                If ipy = my + 1 Then ipy = my
                rx = (xd(id) - xmin) / hx - (ipx - 1)
                ry = (yd(id) - ymin) / hy - (ipy - 1)
                bspl(rx, hx, 0, bx0)
                bspl(rx, hx, 1, bx1)
                bspl(ry, hy, 0, by0)
                bspl(ry, hy, 1, by1)
                fdx = 0.0
                fdy = 0.0

                For j = 0 To 3
                    For i = 0 To 3
                        ij = ipx + i + (ipy - 1 + j) * (mx + 3)
                        fdx = fdx + c(ij) * bx1(4 - i) * by0(4 - j)
                        fdy = fdy + c(ij) * bx0(4 - i) * by1(4 - j)
                    Next
                Next

                dxres = fdx - dfdx(id)
                dyres = fdy - dfdy(id)
                Rdx = Rdx + dxres ^ 2
                Rdy = Rdy + dyres ^ 2
            End If
        Next

        QRd = (Rdx + Rdy) / (nddt)
        QRdav = Math.Sqrt(QRd)

    End Sub

    Private Sub Jvalue(ByVal mx As Integer, ByVal my As Integer, ByVal pm1 As Double, ByVal pm2 As Double, _
                       ByVal QJx As Double, ByVal QJy As Double, ByVal QJxx As Double, ByVal QJxy As Double, ByVal QJyy As Double, _
                       ByVal QJ1 As Double, ByVal QJ2 As Double, ByVal QJ As Double)

        '/////////////////subroutine Jvalue////////////////////////////////////
        '   Evaluate the smoothness of final solution.
        '      J(f) = m1 * J1(f) + m2 * J2(f)
        '           = m1 * [double integral {(fx)**2 + (fy)**2} dxdy]
        '           + m2 * [double integral {(fxx)**2+2*(fxy)**2+(fyy)**2}dxdy]
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      QJx   : double integration of (fx)**2
        '      QJy   : double integration of (fy)**2
        '      QJxx  : double integration of (fxx)**2
        '      QJxy  : double integration of (fxy)**2
        '      QJyy  : double integration of (fyy)**2
        '      QJ1   : value of J1(f)
        '      QJ2   : value of J2(f)
        '      QJ    : value of J(f)
        '//////////////////////////////////////////////////////////////////////

        Dim wx, wy, wxx, wxy, wyy As Double

        Dim ffix0(4), ffix1(4), ffix2(4) As Double
        Dim ffiy0(4), ffiy1(4), ffiy2(4) As Double

        Dim i1, j1, i22, j22 As Long

        wx = hy / hx / (mx * My)
        wy = hx / hy / (mx * My)
        wxx = hy / (hx ^ 3)
        wxy = 1.0 / (hx * hy)
        wyy = hx / (hy ^ 3)
        QJx = 0.0
        QJy = 0.0
        QJxx = 0.0
        QJxy = 0.0
        QJyy = 0.0

        For j1 = 1 To my + 3
            integral(my, 0, j1, ffiy0)
            integral(my, 1, j1, ffiy1)
            integral(my, 2, j1, ffiy2)
            j22 = 4
            If j1 >= my + 1 Then j22 = my + 4 - j1

            For i1 = 1 To mx + 3
                integral(mx, 0, i1, ffix0)
                integral(mx, 1, i1, ffix1)
                integral(mx, 2, i1, ffix2)
                i22 = 4
                If i1 >= mx + 1 Then i22 = mx + 4 - i1

                Jtmp(mx, wx, i1, j1, i22, j22, ffix1, ffiy0, QJx)
                Jtmp(mx, wy, i1, j1, i22, j22, ffix0, ffiy1, QJy)
                Jtmp(mx, wxx, i1, j1, i22, j22, ffix2, ffiy0, QJxx)
                Jtmp(mx, wxy, i1, j1, i22, j22, ffix1, ffiy1, QJxy)
                Jtmp(mx, wyy, i1, j1, i22, j22, ffix0, ffiy2, QJyy)
            Next
        Next
        QJ1 = (QJx + QJy)
        QJ2 = QJxx + 2.0 * QJxy + QJyy
        QJ = pm1 * QJ1 + pm2 * QJ2

    End Sub
    '

    Private Sub Jtmp(ByVal mx As Long, ByVal w As Double, ByVal i1 As Long, ByVal j1 As Long, ByVal i22 As Long, ByVal j22 As Long, _
                     ByVal ffix() As Double, ByVal ffiy() As Double, ByRef QJ As Double)

        '/////////////////subroutine Jtmp//////////////////////////////////////
        '    Calculate the double integration of partial derivative squared
        '    using final coefficients matrix c.
        '       Jx  = ct * Ax  * c
        '       Jy  = ct * Ay  * c
        '       Jxx = ct * Axx * c
        '       Jxy = ct * Axy * c * 2
        '       Jyy = ct * Ayy * c
        '    where ct is an inverse matrix of c.
        '----------------------------------------------------------------------
        '   Arguments used in subroutine
        '      w       : weight of element 
        '      i1,j1   : element position indicator
        '      i22,j22 : indicator where is the limit of copy
        '      QJ      : double integration partial derivative squared
        '//////////////////////////////////////////////////////////////////////

        Dim tmp As Double

        Dim i2, j2, ia1, ja1, ia2, ja2 As Long

        For j2 = 1 To j22
            For i2 = 1 To i22
                tmp = w * ffix(i2) * ffiy(j2)
                ia1 = i1 + (j1 - 1) * (mx + 3)
                ja1 = (i1 + i2 - 1) + ((j1 + j2 - 1) - 1) * (mx + 3)
                If ia1 = ja1 Then
                    QJ = QJ + c(ia1) * tmp * c(ja1)
                Else
                    QJ = QJ + c(ia1) * tmp * c(ja1) * 2.0
                    If j2 <> 1 And i2 <> 1 Then
                        ia2 = (i1 + i2 - 1) + (j1 - 1) * (mx + 3)
                        ja2 = i1 + ((j1 + j2 - 1) - 1) * (mx + 3)
                        QJ = QJ + c(ia2) * tmp * c(ja2) * 2.0
                    End If
                End If
            Next
        Next

    End Sub

End Module
